@@ -3,7 +3,7 @@
  *
  *   Copyright (C) UENO Katsuhiro 2000-2003
  *
- * $Id: zlib.c 19892 2008-10-22 16:53:50Z matz $
+ * $Id: zlib.c 23263 2009-04-22 16:31:06Z knu $
  */
 
 #include <ruby.h>
@@ -163,6 +163,7 @@ static VALUE rb_gzfile_sync(VALUE);
 static VALUE rb_gzfile_set_sync(VALUE, VALUE);
 static VALUE rb_gzfile_total_in(VALUE);
 static VALUE rb_gzfile_total_out(VALUE);
+static VALUE rb_gzfile_path(VALUE);
 
 static VALUE rb_gzwriter_s_allocate(VALUE);
 static VALUE rb_gzwriter_s_open(int, VALUE*, VALUE);
@@ -695,7 +696,7 @@ zstream_run(struct zstream *z, Bytef *src, uInt len, int flush)
 {
     uInt n;
     int err;
-    volatile VALUE guard;
+    volatile VALUE guard = Qnil;
 
     if (NIL_P(z->input) && len == 0) {
 	z->stream.next_in = (Bytef*)"";
@@ -716,6 +717,9 @@ zstream_run(struct zstream *z, Bytef *src, uInt len, int flush)
     }
 
     for (;;) {
+	/* VC allocates err and guard to same address.  accessing err and guard
+	   in same scope prevents it. */
+	RB_GC_GUARD(guard);
 	n = z->stream.avail_out;
 	err = z->func->run(&z->stream, flush);
 	z->buf_filled += n - z->stream.avail_out;
@@ -1650,7 +1654,7 @@ rb_inflate_set_dictionary(VALUE obj, VALUE dic)
 #define OS_CODE  OS_UNIX
 #endif
 
-static ID id_write, id_read, id_readpartial, id_flush, id_seek, id_close;
+static ID id_write, id_read, id_readpartial, id_flush, id_seek, id_close, id_path;
 static VALUE cGzError, cNoFooter, cCRCError, cLengthError;
 
 
@@ -1675,6 +1679,7 @@ struct gzfile {
     int ecflags;
     VALUE ecopts;
     char *cbuf;
+    VALUE path;
 };
 #define GZFILE_CBUF_CAPA 10
 
@@ -1696,6 +1701,7 @@ gzfile_mark(struct gzfile *gz)
     rb_gc_mark(gz->comment);
     zstream_mark(&gz->z);
     rb_gc_mark(gz->ecopts);
+    rb_gc_mark(gz->path);
 }
 
 static void
@@ -1742,6 +1748,7 @@ gzfile_new(klass, funcs, endfunc)
     gz->ecflags = 0;
     gz->ecopts = Qnil;
     gz->cbuf = 0;
+    gz->path = Qnil;
 
     return obj;
 }
@@ -2670,6 +2677,21 @@ rb_gzfile_total_out(VALUE obj)
     return rb_uint2inum(gz->z.stream.total_out - gz->z.buf_filled);
 }
 
+/*
+ * Document-method: path
+ *
+ * call-seq: path
+ *
+ * Returns the path string of the associated IO-like object.  This
+ * method is only defined when the IO-like object responds to #path().
+ */
+static VALUE
+rb_gzfile_path(VALUE obj)
+{
+    struct gzfile *gz;
+    Data_Get_Struct(obj, struct gzfile, gz);
+    return gz->path;
+}
 
 static void
 rb_gzfile_ecopts(struct gzfile *gz, VALUE opts)
@@ -2767,6 +2789,11 @@ rb_gzwriter_initialize(int argc, VALUE *argv, VALUE obj)
     ZSTREAM_READY(&gz->z);
     rb_gzfile_ecopts(gz, opt);
 
+    if (rb_respond_to(io, id_path)) {
+	gz->path = rb_funcall(gz->io, id_path, 0);
+	rb_define_singleton_method(obj, "path", rb_gzfile_path, 0);
+    }
+
     return obj;
 }
 
@@ -2962,6 +2989,11 @@ rb_gzreader_initialize(int argc, VALUE *argv, VALUE obj)
     gzfile_read_header(gz);
     rb_gzfile_ecopts(gz, opt);
 
+    if (rb_respond_to(io, id_path)) {
+	gz->path = rb_funcall(gz->io, id_path, 0);
+	rb_define_singleton_method(obj, "path", rb_gzfile_path, 0);
+    }
+
     return obj;
 }
 
@@ -3415,7 +3447,8 @@ rb_gzreader_readlines(int argc, VALUE *argv, VALUE obj)
  *   Zlib::OS_UNKNOWN
  *       The return values of Zlib::GzipFile#os_code method.
  */
-void Init_zlib()
+void
+Init_zlib()
 {
     VALUE mZlib, cZStream, cDeflate, cInflate;
 #if GZIP_SUPPORT
@@ -3512,6 +3545,7 @@ void Init_zlib()
     id_flush = rb_intern("flush");
     id_seek = rb_intern("seek");
     id_close = rb_intern("close");
+    id_path = rb_intern("path");
 
     cGzipFile = rb_define_class_under(mZlib, "GzipFile", rb_cObject);
     cGzError = rb_define_class_under(cGzipFile, "Error", cZError);
@@ -3571,7 +3605,7 @@ void Init_zlib()
     rb_define_method(cGzipReader, "getc", rb_gzreader_getc, 0);
     rb_define_method(cGzipReader, "getbyte", rb_gzreader_getbyte, 0);
     rb_define_method(cGzipReader, "readchar", rb_gzreader_readchar, 0);
-    rb_define_method(cGzipReader, "readchar", rb_gzreader_readbyte, 0);
+    rb_define_method(cGzipReader, "readbyte", rb_gzreader_readbyte, 0);
     rb_define_method(cGzipReader, "each_byte", rb_gzreader_each_byte, 0);
     rb_define_method(cGzipReader, "each_char", rb_gzreader_each_char, 0);
     rb_define_method(cGzipReader, "bytes", rb_gzreader_each_byte, 0);
